<?php

declare(strict_types=1);

namespace Drupal\Tests\calendar\Functional;

use Drupal\Tests\views\Functional\ViewTestBase;
use Drupal\user\UserInterface;
use Drupal\views\Views;

/**
 * Test description.
 *
 * @group calendar
 */
class GenericTest extends ViewTestBase {

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'block',
    'calendar',
    'datetime',
    'datetime_range',
    'views_ui',
    'node',
    'calendar_module_test',
  ];

  /**
   * Views used by this test.
   *
   * @var string[]
   */
  public static array $testViews = ['content_authored_on_calendar'];

  /**
   * A user with admin privileges.
   */
  protected UserInterface $adminUser;

  /**
   * {@inheritdoc}
   */
  protected function setUp($import_test_views = TRUE, $modules = ['calendar_module_test']): void {
    parent::setUp($import_test_views, $modules);

    $this->drupalPlaceBlock('local_tasks_block');
    $this->drupalPlaceBlock('local_actions_block');

    // Create a user with the 'administer site configuration' permission.
    $this->adminUser = $this->drupalCreateUser([
      'administer calendar settings',
      'administer views',
    ]);
    $this->drupalLogin($this->adminUser);
  }

  /**
   * Tests calendar admin settings page.
   */
  public function testCalendarAdminPage(): void {
    $this->drupalGet('admin/config/date/calendar');
    $session = $this->assertSession();
    $session->statusCodeEquals(200);
    $session->buttonExists('Save configuration')->submit();
    $session->pageTextContains('The configuration options have been saved');
  }

  /**
   * Tests if the 'Add view from template' link exists.
   */
  public function testAddViewFromTemplateLinkExists(): void {
    $this->drupalGet('admin/structure/views');
    $this->assertSession()->statusCodeEquals(200);

    // Retrieve the link and verify it points to the expected URL.
    $session = $this->assertSession();
    $session->linkExists('Add view from template');
  }

  /**
   * Test an event appears in the right cell in the calendar.
   */
  public function testEventAppearsOnCorrectDayInCalendar(): void {
    // Set the date for the node to the first of a fixed date.
    $date = new \DateTime('2025-04-01 12:00:00');

    // Create a node with the specified date.
    $this->drupalCreateContentType(['type' => 'article']);
    $node = $this->drupalCreateNode([
      'type' => 'article',
      'created' => $date->getTimestamp(),
    ]);

    $view = Views::getView('content_authored_on_calendar');
    $view->setDisplay('page_month');
    $view->execute();

    // Navigate to the calendar page.
    $output = $this->drupalGet($view->getUrl([
      $date->format('Ym'),
    ]));

    // Check the header is on the page.
    $this->assertStringContainsString($date->format('F Y'), $output);
    // Ensure the node title is on the page, somewhere.
    $this->assertSession()->pageTextContains($node->label());

    // Locate the calendar cell for the specified date.
    $day_of_month = $date->format('j');
    $xpath = "//tr[contains(@class, 'single-day')]//td[@data-day-of-month='{$day_of_month}']";
    $calendar_cell = $this->getSession()->getPage()->find('xpath', $xpath);

    // Assert that the calendar cell exists.
    $this->assertNotNull($calendar_cell, "The calendar cell for day {$day_of_month} exists.");

    // Check if the node's title is present in the calendar cell.
    $this->assertStringContainsString(
      $node->label(),
      $calendar_cell->getHtml(),
      "The node titled '{$node->label()}' appears in the calendar cell for day {$day_of_month}."
    );
  }

}
